#!/usr/bin/env python

############################################################################
# Copyright (c) 2008,2009,2010,2011 SFB/TR-55
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
# 
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import string
import qlistmod as qq

############################################################################
# constant declarations
############################################################################

fuqdsock = None
timeout = 240

QFC_OUTPUT_BOARDER_SIZE = 110

QFC_FLASH_RETRY_THRESHOLD_INTERNAL = 5
QFC_FLASH_RETRY_THRESHOLD_EXTERNAL = 15
QFC_FLASH_MAX_PARALLEL             = 32
QFC_PCMD_MAX_PARALLEL              = 64

# --------------------------------------------------------------------
# dictionary key definitions
# --------------------------------------------------------------------

# target identifiers
QFC_KEY_NODES       = "nodes"
QFC_KEY_PSUS        = "psus"

# errors identifiers
QFC_KEY_GLOBAL_ERR  = "global errors"
QFC_KEY_LOCAL_ERR   = "local errors"
QFC_KEY_SUCC_LIST   = "success list"
QFC_KEY_FAIL_LIST   = "fail list"

# --------------------------------------------------------------------
# command definitions
# --------------------------------------------------------------------

# get
QFC_CMD_GET_NODE_TEMP       = "get node temp"
QFC_CMD_GET_NODE_SERIAL     = "get node serial"
QFC_CMD_GET_NODE_MAC        = "get node mac"
QFC_CMD_GET_NODE_VPD        = "get node vpd"
QFC_CMD_GET_NODE_STATUS     = "get node status"
QFC_CMD_GET_NODE_CONFIG     = "get node config"
QFC_CMD_GET_NODE_DUMP       = "get node dump"
QFC_CMD_GET_NODE_VERSION    = "get node version"
QFC_CMD_GET_PSU_STATUS      = "get psu status"

# set
QFC_CMD_SET_NODE_MAC        = "set node mac"
QFC_CMD_SET_NODE_VPD        = "set node vpd"
QFC_CMD_SET_NODE_CLOCK      = "set node clock"

# power
QFC_CMD_POWER_NODE          = "power node"
QFC_CMD_POWER_NODE_OFF      = "power node off"
QFC_CMD_POWER_NODE_ON       = "power node on"
QFC_CMD_POWER_NODE_CYCLE    = "power node cycle"
QFC_CMD_POWER_NODE_SP       = "power node sp"
QFC_CMD_POWER_NODE_NOBOOT   = "power node noboot"
QFC_CMD_POWER_NODE_PROBE    = "power node probe"
QFC_CMD_POWER_NODE_RESET    = "power node reset"
QFC_CMD_POWER_NODE_REBOOT   = "power node reboot"
QFC_CMD_POWER_NODE_HALT     = "power node halt"
QFC_CMD_POWER_NODE_MAGIC    = "power node magic"
QFC_CMD_POWER_NODE_WOOF     = "power node woof"
QFC_CMD_POWER_PSU           = "power psu"
QFC_CMD_POWER_PSU_OFF       = "power psu off"
QFC_CMD_POWER_PSU_ON        = "power psu on"

# flash
QFC_CMD_FLASH_NODE_LINUX    = "flash node linux"
QFC_CMD_FLASH_NODE_SP       = "flash node sp"
QFC_CMD_FLASH_NODE_FPGA     = "flash node fpga"
QFC_CMD_FLASH_NODE_SLOF     = "flash node slof"
QFC_CMD_FLASH_NODE_INTERNAL = "flash node internal"
QFC_CMD_FLASH_NODE_EXTERNAL = "flash node external"

# clear
QFC_CMD_CLEAR_NODE_STATES   = "clear node states"

# other
QFC_CMD_LINKSTRESS          = "stress links"

# --------------------------------------------------------------------
# access definitions
# --------------------------------------------------------------------

# service processor commands
QFC_NODE_SP_TEMPERATURE     = 0x46
QFC_NODE_SP_VPD_READ        = 0x31
QFC_NODE_SP_CPLD_READ       = 0x42
QFC_NODE_SP_STATUS_CHANGE   = 0x33
QFC_NODE_SP_SET_NODE_ID     = 0x41
QFC_NODE_SP_VPD_WRITE       = 0x32
QFC_NODE_SP_CPLD_WRITE      = 0x43
QFC_NODE_SP_POWER           = 0x30
QFC_NODE_SP_VERSION         = 0x45
QFC_NODE_SP_CLEAR_STATES    = 0x39

# vpd address and length
QFC_NODE_VPD_BLOCKSIZE       = 250
QFC_NODE_VPD_MAC_ADDR        = 0x0004
QFC_NODE_VPD_MAC_LEN         = 6
QFC_NODE_VPD_SERIAL_ADDR     = 0x7020
QFC_NODE_VPD_SERIAL_LEN      = 22
QFC_NODE_VPD_BE_FIR_ADDR     = 0x2000
QFC_NODE_VPD_BE_FIR_LEN      = 0x300
QFC_NODE_VPD_FPGA_FIR_ADDR   = 0x2600
QFC_NODE_VPD_FPGA_FIR_LEN    = 0x48
QFC_NODE_VPD_FIR_HEADER_ADDR = 0x2000
QFC_NODE_VPD_FIR_HEADER_LEN  = 4

# flash
QFC_NODE_FLASH_BLOCKSIZE         = qq.FEEN_SECTOR_SIZE_FLASH_EXTERNAL
QFC_NODE_FLASH_SLOF_VERSION_ADDR = 0x400038
QFC_NODE_FLASH_SLOF_VERSION_LEN  = 0x10
QFC_NODE_FLASH_FPGA_VERSION_ADDR = 0
QFC_NODE_FLASH_FPGA_VERSION_LEN  = 0x400
QFC_NODE_FLASH_SP_IMAGE_ADDR     = 0
QFC_NODE_FLASH_FPGA_IMAGE_ADDR   = 0
QFC_NODE_FLASH_SLOF_IMAGE_ADDR   = 0x400000

# psu pmbus
QFC_PSU_PMBUS_STATUS_WORD   = 0x79
QFC_PSU_PMBUS_VOUT          = 0x8b
QFC_PSU_PMBUS_VOUT_COMMAND  = 0x21
QFC_PSU_PMBUS_VOUT_VIN      = 0x88
QFC_PSU_PMBUS_VOUT_IOUT     = 0x8c
QFC_PSU_PMBUS_VOUT_IIN      = 0x89
QFC_PSU_PMBUS_VOUT_POUT     = 0x96
QFC_PSU_PMBUS_VOUT_PIN      = 0x97
QFC_PSU_PMBUS_TEMPERATURE_1 = 0x8d
QFC_PSU_PMBUS_TEMPERATURE_2 = 0x8e
QFC_PSU_PMBUS_FAN_SPEED_1   = 0x90
QFC_PSU_PMBUS_FAN_SPEED_2   = 0x91

# --------------------------------------------------------------------
# command method definitions
# --------------------------------------------------------------------

# the method map contains a command to command-method remapping
# this is required for code re-usage
QFC_METHOD_MAP = {}

# get
QFC_METHOD_MAP[QFC_CMD_GET_NODE_TEMP]       = [QFC_CMD_GET_NODE_TEMP, None]
QFC_METHOD_MAP[QFC_CMD_GET_NODE_SERIAL]     = \
    [QFC_CMD_GET_NODE_VPD, [QFC_NODE_VPD_SERIAL_ADDR, QFC_NODE_VPD_SERIAL_LEN, None]]
QFC_METHOD_MAP[QFC_CMD_GET_NODE_MAC]        = \
    [QFC_CMD_GET_NODE_VPD, [QFC_NODE_VPD_MAC_ADDR, QFC_NODE_VPD_MAC_LEN, None]]
QFC_METHOD_MAP[QFC_CMD_GET_NODE_VPD]        = [QFC_CMD_GET_NODE_VPD, None]
QFC_METHOD_MAP[QFC_CMD_GET_NODE_STATUS]     = [QFC_CMD_GET_NODE_STATUS, None]
QFC_METHOD_MAP[QFC_CMD_GET_NODE_CONFIG]     = [QFC_CMD_GET_NODE_CONFIG, None]
QFC_METHOD_MAP[QFC_CMD_GET_NODE_DUMP]       = [QFC_CMD_GET_NODE_DUMP, None]
QFC_METHOD_MAP[QFC_CMD_GET_NODE_VERSION]    = [QFC_CMD_GET_NODE_VERSION, None]
QFC_METHOD_MAP[QFC_CMD_GET_PSU_STATUS]      = [QFC_CMD_GET_PSU_STATUS, None]

# set
QFC_METHOD_MAP[QFC_CMD_SET_NODE_MAC]        = [QFC_CMD_SET_NODE_MAC, None]
QFC_METHOD_MAP[QFC_CMD_SET_NODE_VPD]        = [QFC_CMD_SET_NODE_VPD, None]
QFC_METHOD_MAP[QFC_CMD_SET_NODE_CLOCK]      = [QFC_CMD_SET_NODE_CLOCK, None]

# power
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_OFF]      = [QFC_CMD_POWER_NODE, [None, None, ['00']]]
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_ON]       = [QFC_CMD_POWER_NODE, [None, None, ['01']]]
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_CYCLE]    = [QFC_CMD_POWER_NODE, [None, None, ['02']]]
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_SP]       = [QFC_CMD_POWER_NODE, [None, None, ['03']]]
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_NOBOOT]   = [QFC_CMD_POWER_NODE, [None, None, ['04']]]
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_PROBE]    = [QFC_CMD_POWER_NODE_PROBE, None]
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_RESET]    = [QFC_CMD_POWER_NODE_RESET, None]
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_REBOOT]   = [QFC_CMD_POWER_NODE_REBOOT, None]
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_HALT]     = [QFC_CMD_POWER_NODE_HALT, None]
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_MAGIC]    = None
QFC_METHOD_MAP[QFC_CMD_POWER_NODE_WOOF]     = None
QFC_METHOD_MAP[QFC_CMD_POWER_PSU_OFF]       = [QFC_CMD_POWER_PSU, [None, None, [0]]]
QFC_METHOD_MAP[QFC_CMD_POWER_PSU_ON]        = [QFC_CMD_POWER_PSU, [None, None, [1]]]

# flash
QFC_METHOD_MAP[QFC_CMD_FLASH_NODE_LINUX]    = [QFC_CMD_FLASH_NODE_LINUX, None]
QFC_METHOD_MAP[QFC_CMD_FLASH_NODE_SP]       = [QFC_CMD_FLASH_NODE_INTERNAL, None]
QFC_METHOD_MAP[QFC_CMD_FLASH_NODE_FPGA]     = [QFC_CMD_FLASH_NODE_EXTERNAL, None]
QFC_METHOD_MAP[QFC_CMD_FLASH_NODE_SLOF]     = [QFC_CMD_FLASH_NODE_EXTERNAL, None]

# clear
QFC_METHOD_MAP[QFC_CMD_CLEAR_NODE_STATES]   = [QFC_CMD_CLEAR_NODE_STATES, None]

# --------------------------------------------------------------------
# engine access definitions
# --------------------------------------------------------------------

# register any commands here into the engine list
# any command that is not registered cannot be 
# mapped to the corresponding engine

# pcmd engine
QFC_ENGINE_PCMD_LIST  = [QFC_CMD_POWER_NODE_PROBE, QFC_CMD_POWER_NODE_REBOOT, QFC_CMD_POWER_NODE_HALT,
                         QFC_CMD_FLASH_NODE_LINUX, QFC_CMD_LINKSTRESS]

# flash engine
QFC_ENGINE_FLASH_LIST = [QFC_CMD_FLASH_NODE_SP, QFC_CMD_FLASH_NODE_FPGA, QFC_CMD_FLASH_NODE_SLOF]

# construct fuqd engine commands via method map
QFC_ENGINE_FUQD_LIST  = [cmd for cmd in QFC_METHOD_MAP \
                         if ((cmd not in QFC_ENGINE_PCMD_LIST) and (cmd not in QFC_ENGINE_FLASH_LIST))]

############################################################################
# helper functions
############################################################################

# psu2sroot splits a symbolic psu name
# into the corresponding superrootcard-centric
# addressing scheme via superrootcard-ip and psu slot
def psu2sroot(psu):

    psusplit = string.split(psu, '-')
    bp    = string.atoi(psusplit[1])
    bppos = string.atoi(psusplit[2])
    src   = bp / 8
    slot  = ((bp % 8) * 3) + bppos  # goes from 0..23
    srcip = 'src-%1d' % src
    return srcip, slot

#---------------------------------------------------------------------------
# node2root splits a symbolic nodecard name
# into the corresponding rootcard-centric
# addressing scheme via rootcard-ip and node slot
def node2root(host):

    ncsplit = string.split(host, '-')
    bp = string.atoi(ncsplit[1])
    pos = string.atoi(ncsplit[2])
    rcip = 'rc-%02i-%02i' % (bp, pos/16)
    node = pos % 16 # remove backplane dependency
    return rcip, node

#---------------------------------------------------------------------------
# node2bp splits a symbolic nodecard name
# into the corresponding backplane number
# and node slot
def node2bp(host):

    ncsplit = string.split(host, '-')
    bp  = string.atoi(ncsplit[1])
    pos = string.atoi(ncsplit[2])
    return bp, pos

#---------------------------------------------------------------------------
# binaryStr2qdataStr returns a concatenated string
# readable by qdata
def binaryStr2qdataStr(data):

    s = ''
    for d in data:
        s += '%s ' % (d)
    return s

#---------------------------------------------------------------------------
# binaryStr2int returns a list of
# 1-byte integers constructed from
# a list of binary data encoded in string format
def binaryStr2int(data):

    intlist = [int(i, 16) for i in data]
    return intlist

#---------------------------------------------------------------------------
# binaryStr2quad returns a list of
# 4-byte integers constructed from
# a list of binary data encoded in string format
def binaryStr2quad(data):

    quadlist = []
    strel = len(data) / 4
    for el in range(strel):
        off = 4 * el
        quad = int(data[off] + data[off+1] + data[off+2] + data[off+3], 16)
        quadlist.append(quad)
    return quadlist

#---------------------------------------------------------------------------
# binaryStrPad returns a list of
# binary encoded strings with original data,
# but padded and with completed tail
def binaryStrPad(data, padding, base=16, pattern='xx'):

    padded = [pattern for item in range(padding)]
    padded.extend(data)
    taillength = (16 - (len(padded) % base)) % 16
    padded.extend([pattern for item in range(taillength)])
    return padded

#---------------------------------------------------------------------------
# quad2binary converts 4-byte int into 32 bit representation
def quad2binary(data):

    s = ''
    for i in range(32):
        if ((i % 8) == 0) and (i > 0): s+= ' '
        if (data & (0x80000000 >> i)): s += '1'
        else: s += '0'
    return s

#---------------------------------------------------------------------------
# list2subList splits a given list
# into sub lists, depending on the
# number of desired entries in each
# sub list
def list2subLists(data, size):

    sublists = []
    dataindex = 0
    datalen = len(data)
    numfull, numfrac = divmod(datalen, size)
    if numfull > 0:
        for l in range(numfull):
            sublists.append(data[dataindex:dataindex+size])
            dataindex += size
    if numfrac > 0:
        sublists.append(data[dataindex:])
    return sublists

#---------------------------------------------------------------------------
# extractSingleRootcardNode extracts a single
# node connect to a rootcard from a given list
def extractSingleRootcardNode(list, limit=QFC_FLASH_MAX_PARALLEL):

    rcdict = {}
    list.sort()
    for node in list:
        rc, nc = node2root(node)
        if not rc in rcdict:
            rcdict[rc] = []
        rcdict[rc].append(node)

    rlist = []
    counter = 0
    for rc in rcdict:
        counter += 1
        if counter > limit: break
        rlist.append(rcdict[rc][0])
    return rlist

#---------------------------------------------------------------------------
# chunkList creates a list with data lengths
# for split transactions
def chunkList(datalen, size):

    fullsize, restsize = divmod(datalen, size)
    chunks = [size for item in range(fullsize)]
    if restsize > 0:
        chunks.append(restsize)
    return chunks

############################################################################
# qfch - decoder functions
############################################################################

# ncBEglobalFirDecode decodes the BE global fir
# registers into human readable string :-D
def ncBEglobalFirDecode(code):

    decoded = ''
    if code == 0x00000000: decoded += " fine "
    if code & 0x00000001 : decoded += " Checkstop by Cell BE is quiesced "
    if code & 0x00000002 : decoded += " Any Local Recoverable Error Counter "
    if code & 0x00000004 : decoded += " Checkstop on trigger "
    if code & 0x00000008 : decoded += " External Checkstop C4 pin "
    if code & 0x00000010 : decoded += " SPE 7 "
    if code & 0x00000020 : decoded += " SPE 6 "
    if code & 0x00000040 : decoded += " SPE 5 "
    if code & 0x00000080 : decoded += " SPE 4 "
    if code & 0x00000100 : decoded += " SPE 3 "
    if code & 0x00000200 : decoded += " SPE 2 "
    if code & 0x00000400 : decoded += " SPE 1 "
    if code & 0x00000800 : decoded += " SPE 0 "
    if code & 0x00001000 : decoded += " BIC/IOC "
    if code & 0x00002000 : decoded += " MIC "
    if code & 0x00004000 : decoded += " CIU "
    if code & 0x00008000 : decoded += " BIU "
    if code & 0x00010000 : decoded += " L2-Cache "
    if code & 0x00020000 : decoded += " PPE-PPU "
    if code == 0xffffffff: decoded =  " invalid "
    return decoded                        

#---------------------------------------------------------------------------
# ncBErecoverableFirDecode decodes the BE global fir
# registers into human readable string :-D
def ncBErecoverableFirDecode(code):

    decoded = ''
    if code == 0x00000000: decoded += " fine "
    if code & 0x00000002 : decoded += " SPE 7 "
    if code & 0x00000004 : decoded += " SPE 6 "
    if code & 0x00000008 : decoded += " SPE 5 "
    if code & 0x00000010 : decoded += " SPE 4 "
    if code & 0x00000020 : decoded += " SPE 3 "
    if code & 0x00000040 : decoded += " SPE 2 "
    if code & 0x00000080 : decoded += " SPE 1 "
    if code & 0x00000100 : decoded += " SPE 0 "
    if code & 0x00000200 : decoded += " BIC/IOC " 
    if code & 0x00000400 : decoded += " MIC "     
    if code & 0x00000800 : decoded += " CIU "     
    if code & 0x00001000 : decoded += " BIU "
    if code & 0x00002000 : decoded += " L2-Cache "
    if code & 0x00004000 : decoded += " PPE-PPU "
    if code == 0xffffffff: decoded =  " invalid "
    return decoded                        
                        
#---------------------------------------------------------------------------
# ncPorDecode decodes the POR status into
# human readable string :-D
def ncPorDecode(code):

    if   code == 0x00 : decoded = "no error detected"
    elif code == 0x02 : decoded = "por_cpu0_attn state 0"
    elif code == 0x03 : decoded = "por_cpu0_attn state 1"
    elif code == 0x04 : decoded = "por_uart_req"         
    elif code == 0x05 : decoded = "por_cpu0_spitest"     
    elif code == 0x06 : decoded = "por_sb0_spitest"      
    elif code == 0x21 : decoded = "eqi_prestart"         
    elif code == 0x22 : decoded = "eqi_prepower_off"     
    elif code == 0x23 : decoded = "eqi_postpower_off"    
    elif code == 0x24 : decoded = "eqi_prehreset_active" 
    elif code == 0x25 : decoded = "eqi_posthreset_active"
    elif code == 0x26 : decoded = "eqi_prepower_on"      
    elif code == 0x27 : decoded = "eqi_postpower_on"     
    elif code == 0x28 : decoded = "eqi_prexclocks_on"    
    elif code == 0x29 : decoded = "eqi_postxclocks_on"   
    elif code == 0x2a : decoded = "eqi_prehreset_release"
    elif code == 0x2b : decoded = "eqi_posthreset_release"
    elif code == 0x2c : decoded = "eqi_startup_sync" 
    elif code == 0x2d : decoded = "eqi_end"
    elif code == 0x2e : decoded = "eqi_normal_config"
    elif code == 0x2f : decoded = "eqi_debug_config" 
    elif code == 0x30 : decoded = "eqi_posttrac" 
    elif code == 0x31 : decoded = "eqi_pin_setting"
    elif code == 0x32 : decoded = "eqi_postcfg"
    elif code == 0x33 : decoded = "eqi_prefirmware"
    elif code == 0x40 : decoded = "rrac_cpu0_bx_rdy" 
    elif code == 0x41 : decoded = "rrac_cpu0_bx_cal" 
    elif code == 0x42 : decoded = "rrac_sb0_bx_cal"
    elif code == 0x43 : decoded = "rrac_cpu0_tx_lock"
    elif code == 0x44 : decoded = "rrac_cpu0_rc_lock"
    elif code == 0x45 : decoded = "rrac_cpu0_phase_cal"
    elif code == 0x46 : decoded = "rrac_sb0_bx_rdy"  
    elif code == 0x47 : decoded = "rrac_cpu0_par_cal"
    elif code == 0x48 : decoded = "rrac_sb0_par_cal" 
    elif code == 0x49 : decoded = "rrac_cpu0_lev_cal"
    elif code == 0x4a : decoded = "rrac_sb0_lev_cal" 
    elif code == 0x4b : decoded = "rrac_cpu0_byte_train"
    elif code == 0x4c : decoded = "rrac_sb0_byte_train" 
    else              : decoded = "unknown error code"  
    r = '(%s)' % (decoded)
    return r

#---------------------------------------------------------------------------
# ncStatusEventDecode decodes the event status into
# human readable string :-D
def ncStatusEventDecode(code):

    if   code == 0x00 : decoded = "no event detected" 
    elif code == 0x01 : decoded = "SP boot" 
    elif code == 0x02 : decoded = "New boot counter"
    elif code == 0x03 : decoded = "Boot status"
    elif code == 0x04 : decoded = "CPU fault"
    elif code == 0x05 : decoded = "DIMM fault"
    elif code == 0x06 : decoded = "VDD"
    elif code == 0x07 : decoded = "overtemp"
    elif code == 0x08 : decoded = "thermal alert"
    elif code == 0x09 : decoded = "checkstop"
    elif code == 0x0c : decoded = "watchdog expired"
    elif code == 0x0d : decoded = "power off"
    elif code == 0x0e : decoded = "text message"
    elif code == 0x10 : decoded = "host checkpoint"
    elif code == 0x11 : decoded = "POR error code"
    elif code == 0x12 : decoded = "failed boot retries"
    elif code == 0x13 : decoded = "recoverable FIR"
    elif code == 0x14 : decoded = "CPU attention"
    elif code == 0x15 : decoded = "FPGA check"
    else              : decoded = "unknown event code"
    r = '(%s)' % (decoded)
    return r

#---------------------------------------------------------------------------
# ncBootStatusDecode decodes the boot status into
# human readable string :-D
def ncBootStatusDecode(code):

    if   code == 0xff : decoded = "off"
    elif code == 0xfe : decoded = "critical error"
    elif code == 0x02 : decoded = "slof started"
    elif code == 0x01 : decoded = "slof running"         
    elif code == 0x00 : decoded = "linux running"
    else              : decoded = "unknown state"
    r = '(%s)' % (decoded)
    return r

#---------------------------------------------------------------------------
# ncStatusDecode decodes the status into
# human readable string :-D
def ncStatusDecode(code):

    decoded = ''
    r = ''
    if code & 0x01 : decoded += " be_attention "
    if code & 0x02 : decoded += " be_chkstop_out "
    if code & 0x04 : decoded += " be_pll_lock "
    if code & 0x08 : decoded += " soma0_therl_alert "
    if code & 0x10 : decoded += " soma_over_temp "
    if code & 0x20 : decoded += " fpga_chkstop_out "
    if code & 0x80 : decoded += " fpga_done "
    if len(decoded) > 0:
        r = '[%s]' % (decoded)
    return r

#---------------------------------------------------------------------------
# ncPowerSeqStatusDecode decodes the power sequence status into
# human readable string :-D
def ncPowerSeqStatusDecode(code):

    decoded = ''
    r = ''
    if code & 0x01 : decoded += " pwr_seq0_ok " 
    if code & 0x02 : decoded += " pwr_seq1_ok " 
    if code & 0x04 : decoded += " pwr_seq2_ok " 
    if code & 0x08 : decoded += " pwr_seq3_ok " 
    if len(decoded) > 0:
        r = '[%s]' % (decoded)
    return r

#---------------------------------------------------------------------------
# ncPowerSeqConfigDecode decodes the power sequence config into
# human readable string :-D
def ncPowerSeqConfigDecode(code):

    decoded = ''
    r = ''
    if code & 0x01 : decoded += " pwr_seq0_on " 
    if code & 0x02 : decoded += " pwr_seq1_on " 
    if code & 0x04 : decoded += " pwr_seq2_on " 
    if code & 0x08 : decoded += " pwr_seq3_on " 
    if code & 0x10 : decoded += " emergency_detection " 
    if len(decoded) > 0:
        r = '[%s]' % (decoded)
    return r

#---------------------------------------------------------------------------
# ncBootFlagDecode decodes the boot flags into
# human readable string :-D
def ncBootFlagDecode(code):

    decoded = []
    decoded.append(" normal boot ")
    decoded.append(" 1 byte RRAC ")
    decoded.append(" manual boot ")
    if code & 0x01 : decoded[0] = " debug boot " 
    if code & 0x02 : decoded[1] = " 2 byte RRAC " 
    if code & 0x04 : decoded[2] = " auto boot " 
    r = '[%s%s%s]' % (decoded[0], decoded[1], decoded[2])
    return r

#---------------------------------------------------------------------------
# ncRracClkDecode decodes the RRAC clk definition into
# human readable string :-D
def ncRracClkDecode(code):

    decoded = 'unknown freq'
    if code == 0x02 : decoded = "200 MHz" 
    if code == 0x19 : decoded = "250 MHz" 
    if code == 0x03 : decoded = "300 MHz" 
    r = '(%s)' % (decoded)
    return r

#---------------------------------------------------------------------------
# ncSomaVddStatusDecode decodes the Soma VDD status into
# human readable string :-D
def ncSomaVddStatusDecode(code):

    decoded = ''
    r = ''
    if code & 0x01 : decoded += " pwr_input_oc "
    if code & 0x02 : decoded += " som0vdd_fault "
    if code & 0x04 : decoded += " som0vdd_vr_fan "
    if code & 0x08 : decoded += " som0vdd_vr_hot "
    if code & 0x10 : decoded += " som0vdd_ovp_out "
    if code & 0x20 : decoded += " som0vdd_slv_fault "
    if len(decoded) > 0:
        r = '[%s]' % (decoded)
    return r

#---------------------------------------------------------------------------
# psuStatusDecode decodes the psu status word into
# human readable string :-D
def psuStatusDecode(code):

    decoded = ''
    r = ''
    if code & 0x0001 : decoded += " none "
    if code & 0x0002 : decoded += " cml_event "
    if code & 0x0004 : decoded += " temp_fault "
    if code & 0x0008 : decoded += " vin_uv_fault "
    if code & 0x0010 : decoded += " iout_oc_fault "
    if code & 0x0020 : decoded += " vout_ov_fault "
    if code & 0x0040 : decoded += " off "
    if code & 0x0080 : decoded += " busy "
    if code & 0x0100 : decoded += " unknown_fault "
    if code & 0x0200 : decoded += " other_fault "
    if code & 0x0400 : decoded += " fan_fault "
    if code & 0x0800 : decoded += " power_not_good "
    if code & 0x1000 : decoded += " mfr_spec_fault "
    if code & 0x2000 : decoded += " inp_fault "
    if code & 0x4000 : decoded += " iout_fault "
    if code & 0x8000 : decoded += " vout_fault "
    if len(decoded) > 0:
        r = '[%s]' % (decoded)
    return r

#---------------------------------------------------------------------------
# qdataStringify converts qdata into string
def qdataStringify(data):

    str = ''
    for s in data:
        x = int(s, 16)
        if not x: break
        str += '%c' % chr(x)
    return str

#---------------------------------------------------------------------------
# fpgaSectionData returns
# key, data encoded into the bitstream header
def fpgaSectionStr(section, data):
   
    # Thilo's bitstream decoder
    ptr = 13
    for i in range(section+1):
        if ptr >= len(data):
            return None, []
        key = chr(int(data[ptr], 16))
        ptr += 1
        if key == 'e':
            length = binaryStr2quad(data[ptr:ptr+4])[0]
            ptr += 4
        else:
            length = (int(data[ptr], 16) << 8) + int(data[ptr+1], 16)
            ptr += 2
        d = ptr
        ptr += length
    #print 'section', section,' data', qdataStringify(data[d:ptr])
    return key, data[d:ptr]

#---------------------------------------------------------------------------
# fpgaIdentifyBitstream reconstructs the custom stream tag
def fpgaIdentifyBitstream(data):

    # check header
    header = [0x0, 0x9, 0xf, 0xf0, 0xf, 0xf0, 0x0f, 0xf0, 0xf, 0xf0, 0x0, 0x0, 0x1]
    if header != binaryStr2int(data[0:len(header)]):
        return 'bad header'
    section = []
    for s in range(6):
        key, str = fpgaSectionStr(s, data)
        if not key: break
        #print s, key, qdataStringify(str)
        section.append(qdataStringify(str))
    if len(section) < 4: return 'error decoding header, missing sections'
    nwprev = '%s;%s,%s' % (section[0], section[2], section[3])
    return nwprev
    

